/*************************************************************************
 *                                                                       *
 *                    \  `~-'   `-._                                     *
 *                     \\           `.                                   *
 *                     /:     :    /  \                                  * 
 *                    { |  ..'.`  / \  \                                 *
 *                     "|     :  /   `. \                                *
 *                      |    :  /     ) :                                *
 *                      ).   : (     / ,'                                *
 *                     /        `.  / /                                  *
 *                    /       Y   \','                                   *
 *                  ,'        )   |_/                                    *
 *                ,'         /  .,'                                      *
 *              ,'       _,"' '';                                        *
 *             /     _..';     /                                         *
 *            (    -'    ;    /                                          *
 *                                                                       *     
 *      ass: Automagical Sudoku Solver                                   *    
 *                                                                       *   
 *                                                                       *  
 * Author: Ziad van Beek                                                 * 
 * Copying Policy: Do whatever you want with this source, as long as     *
 *                 you don't pretend it's yours                          *
 *                                                                       *
 * Revision:                                                             *
 *      0    04/01/2007      initial version                             *
 *      1.0  05/01/2007      puzzle is read from stdin                   *
 *      1.1  12/01/2007      naked  pairs detection                      *        
 *      1.2  14/01/2007      hidden pairs detection                      *
 *      1.3  16/01/2007      improvement of pairs detection              *
 *                           added statistics                            *
 *      1.4  19/01/2007      added locked box/row/col detection          *
 *                           added debug levels                          *
 *      1.41 22/01/2007      debugging of rc_box()                       *                  
 *      1.5  02/02/2007      added x-wing detection                      *
 *                           added ncurses progress indication           *
 *      1.51                 added vertical   x-wing detection           *
 *      1.52 08/02/2007      added horizontal x-wing detection           *
 *      1.6  16/02/2007      partially added swordfish detection         *
 *      1.7                  added xy-wing detection                     *
 *************************************************************************/      
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <stdlib.h>
#include <ncurses.h>

#define VERSION "1.7"
#ifndef BUILD
    #define BUILD 0
#endif

#define HORI    0
#define VERT    1
#define BLOCK   2

struct pair_struct { int pair[ 9 ];
                     int x[ 9 ]; 
                     int y[ 9 ]; 
                     int total_pairs; 
                     int is_naked; 
                     int is_hidden; 
                     int max_pair;
                   } ;

struct box_row_column { int candidate;
                        int x[ 9 ];
                        int y[ 9 ];
                        int total_found;
                      } ; 

struct swordfish_struct { int candidate; 
                          int x[ 9 ];
                          int y[ 9 ];
                        } ;

 
int i_puzzle[9][9][10];
int debug = 0;
int do_show = 0;

// for statistics
int used_singles = 0;
int used_rc_box = 0;
extern int used_x_wing[2]; 
extern int used_swordfish[2]; 
int used_hidden_pairs_detection = 0;
int used_naked_pairs_detection = 0;

void fill_i_puzzle( void );
void stats( int, int, char * );
void print_puzzle( void );
int  next_cell( int *, int *, int );
int  at_last_element( int, int, int );
int  which_box( int, int );
char *which_dir ( int );
void dump_rc( int, int, int );
void show_candidate ( int ) ;
void check_possibilities_s( void );  // singles
void rc_box ( void );
extern void x_wing ( void );
extern void swordfish ( void );
void generate_pairs( int );
void check_naked_pairs( struct pair_struct );  
void check_hidden_pairs( struct pair_struct );  
void delete_pair_elements( struct pair_struct, int );
int  elim_possibility( int, int, int );
void search_uniques( void );
void search_determined( void );
int  puzzle_complete( void );


void stats ( int sec, int usec, char *str_tot_loops ) 
{
    int x, y;
    int found = 0;
    float found_percent;
    float duration;

    for ( x = 0; x < 9; x++ ) {
        for ( y = 0; y < 9; y++ ) {
            if ( i_puzzle[x][y][0] != 0 ) {
                found++;
            }
        }
    }
    found_percent = ((float) ( found ) / 81 ) * 100.0;
    duration = sec + (float) ( usec / 1000000.0 );
    printf ("Solved %d/81 cells (%.1f%%) in %s loops in %.4f seconds.\n", found, found_percent, str_tot_loops, duration );
    printf     ("Techniques used: Singles Detection (%d)\n", used_singles );
    if ( used_rc_box > 0 )
        printf ("                 Locked Candidates (%d)\n", used_rc_box );
    if ( used_hidden_pairs_detection > 0 )
        printf ("                 Hidden Pairs Detection (%d)\n", used_hidden_pairs_detection );
    if ( used_naked_pairs_detection > 0 )
        printf ("                 Naked Pairs Detection (%d)\n", used_naked_pairs_detection );
    if ( used_x_wing[ VERT ] > 0 )
        printf ("                 X-Wing[%s] (%d)\n", which_dir( VERT ), used_x_wing[ VERT ] );
    if ( used_x_wing[ HORI ] > 0 )
        printf ("                 X-Wing[%s] (%d)\n", which_dir( HORI ), used_x_wing[ HORI ] );
    if ( used_swordfish[ VERT ] > 0 )
        printf ("                 Swordfish[%s] (%d)\n", which_dir( VERT ), used_swordfish[ VERT ] );
    if ( used_swordfish[ HORI ] > 0 )
        printf ("                 Swordfish[%s] (%d)\n", which_dir( HORI ), used_swordfish[ HORI ] );

}
    


/*********************************************************************
 * fill_i_puzzle(): copy the unsolved, human readable puzzle from    *
 * stdin into an array for further use.                              * 
 * *******************************************************************/

void fill_i_puzzle(void)
{
    int x, y, z;
    int max_len = 17;
    int i = 0;
    int known = 0;
    char c_puzzle[9][max_len];
    char buf[max_len + 3];
  
    while ( i < 9 ) {
        fgets(buf, sizeof(buf), stdin);
        if (buf[0] != '-') {
            buf[max_len] = '\0';
            strncpy(c_puzzle[i], buf, max_len);
            i++;
        }
    }

    for ( y = 0; y < 9; y++ ) {
        for ( x = 0; x < max_len; x += 2 ) {
            i_puzzle[x/2][y][0] = ( c_puzzle[y][x] - 48 );
            if ( i_puzzle[x/2][y][0] == 0 ) {   // unknown field, fill it with all possibilities
                for ( z = 1; z < 10; z++ ) {
                    i_puzzle[x/2][y][z] = z;
                }
            }
            else {
                known++;
            }
        }
    }
    if ( do_show == 0 ) {
        printf ("Knowns: %d\n", known);
        puts("This is what the unsolved puzzle looks like:");
        print_puzzle();
    }
}



void print_puzzle(void) 
{
    int x, y, x_p, y_p, sleep_time;

    if ( do_show == 1 ) {  
         sleep_time = 100000 + ( rand() % 500000 ) ; // randomize sleep time
         usleep ( sleep_time );
         for ( y = y_p = 0; y < 9; y++, y_p++ ) {
            if (( y == 3 ) || ( y == 6 )) {
                mvaddstr( y_p, 0, "---------------------");
                y_p++;
            }
            for ( x = x_p = 0; x < 9; x++, x_p += 2 ) {
                if (( x == 3 ) || ( x == 6 )) {
                    mvaddch ( y_p, x_p, '|' );
                    x_p += 2;
                }
                if ( i_puzzle[x][y][0] == 0 ) {
                    mvaddch( y_p, x_p, ' ' );
                }
                else {
                    mvaddch( y_p, x_p, ( i_puzzle[x][y][0] + 48 ) );
                }
            }
        }
        refresh();
    }
    else {
        for ( y = 0; y < 9; y++ ) {
            if (( y == 3 ) || ( y == 6 ))
                puts("---------------------");
            for ( x = 0; x < 9; x++ ) {
                if (( x == 3 ) || ( x == 6 )) {
                    printf("| ");
                }
                printf("%d ", i_puzzle[x][y][0]);
            }
            putchar('\n');
        }
        putchar('\n');
    }
}


/***********************************************************************
 * next_cell(): find next cell to look at                              *
 * returns                                                             * 
 *       0 there is no next cell                                       *
 *       1 all other possibilities                                     *
 ***********************************************************************/

int next_cell ( int *x, int *y, int dir )
{
    if ( dir == HORI ) {
        if ( *x == 8 ) {
            if ( *y == 8 ) {
                return ( 0 );
            }
            else {
                *x = 0;
                (*y)++;
            }
        }
        else {
            (*x)++;
        }
    }
    else if ( dir == VERT ) {
        if ( *y == 8 ) {
            if ( *x == 8 ) {
                return ( 0 );
            }   
            else {
                (*x)++;
                *y = 0;
            }   
        }
        else {
            (*y)++;
        }   
    }
    else if ( dir == BLOCK ) {
        if (( *x % 3 == 2 )) {
            if (( *y % 3 == 2 )) {
                if ( *x == 8 ) {
                    if ( *y == 8 ) {
                        return ( 0 );
                    }
                    else {
                        *x -= 8;
                        (*y)++;
                    }
                }
                else {
                    (*x)++;
                    *y -= 2;
                }
            }
            else {
                (*y)++;
                (*x) -= 2;
            }
        }
        else {
            (*x)++;
        }
    }
    else {
         puts ("next_cell(): You've found a bug!");
    }    

    return( 1 );
}



int at_last_element( int x, int y, int dir )
{
    if (( dir == HORI ) && ( x == 8 )) 
        return( 1 );
    else if (( dir == VERT ) && ( y == 8 )) 
        return( 1 );
    else if ( ( dir == BLOCK ) && (( x % 3 == 2) && ( y % 3 == 2)) )   
        return( 1 );
    else
        return( 0 );
}

// which_box(): based on the x and y coordinates, return the box number starting at 0
int which_box ( int x, int y )
{
    int x_box, y_box;

    x_box = x / 3;
    y_box = y / 3;
    return ( x_box + ( y_box * 3 ) );

}


char *which_dir ( int dir )
{
    if ( dir == HORI )
        return( "HORI" );
    else if ( dir == VERT )
        return( "VERT" );
    else if ( dir == BLOCK )
        return( "BLOCK" );
    else 
        return( "unknown (bug)" );

}



void dump_rc ( int x, int y, int dir ) 
{
    int z;

    if ( dir == HORI ) {
        x = 0;
    }
    else if ( dir == VERT ) {
        y = 0;
    }
    else if ( dir == BLOCK ) {
        x -= x % 3;
        y -= y % 3;
    }
    
    printf ("dump_rc: (%d, %d) %s\n", x, y, which_dir ( dir ) );
             
    while (( x < 9 ) && ( y < 9 )) {
        printf( "(%d, %d) ", x, y );
        for ( z = 0; z < 10; z++ ) {
            printf ( "%d ", i_puzzle[x][y][z]);
        }
        putchar( '\n' );
        if ( dir == HORI ) {
            x++;
        }
        else if ( dir == VERT ) {
            y++;    
        }
        else if ( dir == BLOCK ) {
            if (( x % 3 ) == 2 ) {
                if (( y % 3 ) == 2 ) { // we're at the end of the block
                    break;
                }
                else {
                    x -= 2;  // go back to forst col of this block
                    y++; // next col
                }
            }
            else {
                x++; // next row
            }
        }
    }
    puts ("--------");

}



void show_candidate ( int candidate ) 
{
    int x, y, z, found;
    
    printf ("Showing locations for candidate %d\n", candidate );
    for ( y = 0; y < 9; y++ ) {
        if (( y == 3 ) || ( y == 6 ))
            puts("---------------------");
        for ( x = 0; x < 9; x++ ) {
            found = 0;
            for ( z = 1; z < 9; z++ ) {
                if ( i_puzzle[x][y][z] == candidate ) {
                    found = candidate;
                }
            }
            if (( x == 3 ) || ( x == 6 )) {
                printf("| ");
            }           
            printf("%d ", found );
        }
        putchar('\n');
    }  
    putchar('\n');
    
}



void check_possibilities_s(void) 
{
    int i, x, y, x_s, y_s, x_sq, y_sq;
    short skip_rest;
    for ( x = 0; x < 9; x++ ) {
        for ( y = 0; y < 9; y++ ) {
            if ( i_puzzle[x][y][0] == 0 ) { // 0 means that this square needs investigating
                for ( i = 1; i <= 9; i++ ) {
                    skip_rest = 0;
                    /* search horizontally */
                    if ( skip_rest == 0 ) {
                        for ( x_s = 0; x_s < 9; x_s++ ) {
                            if ( x_s != x ) {  // skip otherwise
                                if ( i_puzzle[x_s][y][0] == i ) {
                                    if ( debug > 3 ) {
                                        puts( "check_possibilties(): calling elim_possibility()" );
                                    }
                                    used_singles += elim_possibility(x, y, i);
                                    skip_rest = 1;
                                    break;  // don't waste resources
                                }
                            }
                        }
                    }
                    /* search vertically */
                    if ( skip_rest == 0 ) {
                        for ( y_s = 0; y_s < 9; y_s++ ) {
                            if ( y_s != y ) {  // skip otherwise
                                if ( i_puzzle[x][y_s][0] == i ) {
                                    if ( debug > 3 ) {
                                        puts( "check_possibilties(): calling elim_possibility()" );
                                    }
                                    used_singles += elim_possibility(x, y, i);
                                    skip_rest = 1;
                                    break;  // don't waste resources
                                }
                            }
                        }
                    }
                    /* search within 3x3 square */
                    if ( skip_rest == 0 ) {
                        // determine which 3x3 square were looking at anyway 
                        x_sq = ( x / 3 );
                        y_sq = ( y / 3 );
                        for ( x_s = ( x_sq * 3 ); x_s < (( (x_sq + 1) * 3 ) ); x_s++ ) {
                            for ( y_s = ( y_sq * 3 ); y_s < (( (y_sq + 1) * 3 ) ); y_s++ ) {
                                if (( x_s != x ) && ( y_s != y )) { // otherwise skip 
                                    if ( i_puzzle[x_s][y_s][0] == i ) {
                                        if ( debug > 3 ) {
                                            puts( "check_possibilties(): calling elim_possibility()" );
                                        }
                                        used_singles += elim_possibility(x, y, i);
                                        break;  // don't waste resources
                                    }
                                }
                            }
                        }
                    } 
                }
            }
        }
    }
}




/***************************************************************
 * r/c-box detection                                           *
 ***************************************************************/

void rc_box ( void )
{
    int x, y, z, i, dir, found, block_tmp, rc_tmp, x_s, y_s, box_lim[2];
    struct box_row_column s_box_rc;

    for ( dir = HORI; dir <= BLOCK; dir++ ) {
        check_possibilities_s();
        for ( s_box_rc.candidate = 1; s_box_rc.candidate <= 9; s_box_rc.candidate++ ) {
            s_box_rc.total_found = 0;
            x = 0;
            y = 0;
            
            do {
                if ( i_puzzle[x][y][0] == 0 ) {
                    for ( z = 1;  z < 10; z++ ) {
                        if ( i_puzzle[x][y][z] == s_box_rc.candidate ) {
                            s_box_rc.x[ s_box_rc.total_found ] = x;
                            s_box_rc.y[ s_box_rc.total_found ] = y;
                            s_box_rc.total_found++;
                        }
                    }
                }            
                if ( at_last_element( x, y, dir ) == 1 ) {
                    if (( s_box_rc.total_found <= 3 ) && ( s_box_rc.total_found > 1 )) {
                        if (( dir == HORI ) || ( dir == VERT )) {
                            block_tmp = -1;
                            found = 1;
                            for ( i = 0; i < s_box_rc.total_found; i++ ) {
                                if ( block_tmp == -1 ) {
                                    block_tmp = which_box ( s_box_rc.x[ i ], s_box_rc.y[ i ] );
                                }
                                else {
                                    if ( block_tmp != ( which_box ( s_box_rc.x[ i ], s_box_rc.y[ i ] ) )) {
                                        found = 0;
                                    }
                                }
                                if ( debug > 3 ) {
                                    printf("box_rc(): tot_found = %d block_tmp = %d, which_block = %d x = %d, y = %d cand = %d at (%d, %d) \n",  s_box_rc.total_found, block_tmp, which_box ( s_box_rc.x[ i ], s_box_rc.y[ i ] ),x , y, s_box_rc.candidate, s_box_rc.x[i], s_box_rc.y[i] );
                                }
                                
                            }
                            if ( found == 1 ) {
                                if ( debug > 1 ) {
                                    printf ("Found a locked candidate %d in block %d. Direction was %s\n", s_box_rc.candidate, block_tmp, which_dir( dir ) );
                                }
                                if ( debug > 2 ) {
                                    dump_rc ( s_box_rc.x[ 0 ], s_box_rc.y[ 0 ], BLOCK );
                                    dump_rc ( s_box_rc.x[ 0 ], s_box_rc.y[ 0 ], dir );
                                }
                                
                                /* now delete all candidates from this block except the ones that are on this row or col */
                                x_s = s_box_rc.x[0] - s_box_rc.x[0] % 3;
                                y_s = s_box_rc.y[0] - s_box_rc.y[0] % 3;
 
                                while ( 1 ) {
                                    if (( dir == HORI ) && ( y_s != y )) {
                                        used_rc_box += elim_possibility ( x_s, y_s, s_box_rc.candidate );
                                    }
                                    else if (( dir == VERT ) && ( x_s != x )) {
                                        used_rc_box += elim_possibility ( x_s, y_s, s_box_rc.candidate );
                                    }
                                    
                                    if (( x_s % 3 ) == 2 ) {
                                        if (( y_s % 3 ) == 2 ) { // we're at the end of the block
                                            break;
                                        }
                                        else {
                                            x_s -= 2;  // go back to forst col of this block
                                            y_s++; // next col
                                        }
                                    }
                                    else {
                                        x_s++; // next row
                                    }
                                }
                            }
                        }
                        else if ( dir == BLOCK ) {  // is candidate in only one row / col?
                            //  vertically 
                            rc_tmp = -1;
                            found = 1;
                            box_lim[0] = 10;
                            box_lim[1] = -1;
                            for ( i = 0; i < s_box_rc.total_found; i++ ) {
                                if ( rc_tmp == -1 ) {
                                    rc_tmp = s_box_rc.x[ i ];
                                }
                                else {
                                    if ( rc_tmp != s_box_rc.x[ i ] ) {
                                        found = 0;
                                    }
                                }
                                // find limits
                                if ( s_box_rc.y[i] < box_lim[0] ) {
                                    box_lim[0] = s_box_rc.y[i];
                                }
                                if ( s_box_rc.y[i] > box_lim[1] ) {
                                    box_lim[1] = s_box_rc.y[i];
                                }
                                if ( debug > 3 ) {
                                    printf("box_rc(): tot_found = %d rc_tmp = %d, which_block = %d x = %d, y = %d cand = %d at (%d, %d) box_lim1 = %d box_lim2 = %d\n",  s_box_rc.total_found, rc_tmp, which_box ( s_box_rc.x[ i ], s_box_rc.y[ i ] ),x , y, s_box_rc.candidate, s_box_rc.x[i], s_box_rc.y[i], box_lim[0], box_lim[1] );                            
                                }
                            }
                            if ( found == 1 ) {
                                if ( debug > 1 ) {
                                    printf ("Found a locked candidate %d in col %d. Direction was %s\n", s_box_rc.candidate, rc_tmp, which_dir( dir ) );
                                }
                                if ( debug > 2 ) {
                                    dump_rc ( s_box_rc.x[ 0 ], s_box_rc.y[ 0 ], VERT );
                                    dump_rc ( s_box_rc.x[ 0 ], s_box_rc.y[ 0 ], dir );
                                }
                                
                                x_s = s_box_rc.x[0];
                                for ( y_s = 0; y_s < 9; y_s++ ) {  // delete this candidate from this col except for the candidates in box
                                    if (( y_s < box_lim[0] ) || ( y_s > box_lim[1] )) { 
                                        used_rc_box += elim_possibility ( x_s, y_s, s_box_rc.candidate );
                                    }
                                }
                            }
                            
                            // horizontally
                            rc_tmp = -1;
                            found = 1;
                            box_lim[0] = 10;
                            box_lim[1] = -1;
 
                            for ( i = 0; i < s_box_rc.total_found; i++ ) { 
                                if ( rc_tmp == -1 ) {
                                    rc_tmp = s_box_rc.y[ i ];
                                }
                                else {
                                    if ( rc_tmp != s_box_rc.y[ i ] ) {
                                        found = 0;
                                    }
                                }
                                if ( s_box_rc.x[i] < box_lim[0] ) {
                                    box_lim[0] = s_box_rc.x[i];
                                }
                                if ( s_box_rc.x[i] > box_lim[1] ) {
                                    box_lim[1] = s_box_rc.x[i];
                                }
                                if ( debug > 3 ) {
                                    printf("box_rc(): tot_found = %d rc_tmp = %d, which_block = %d x = %d, y = %d cand = %d at (%d, %d) box_lim1 = %d box_lim2 = %d\n",  s_box_rc.total_found, rc_tmp, which_box ( s_box_rc.x[ i ], s_box_rc.y[ i ] ),x , y, s_box_rc.candidate, s_box_rc.x[i], s_box_rc.y[i], box_lim[0], box_lim[1] );
                                }
                            }
                            if ( found == 1 ) {
                                if ( debug > 1 ) {
                                    printf ("Found a locked candidate %d in row %d. Direction was %s\n", s_box_rc.candidate, rc_tmp, which_dir( dir ) );
                                }
                                if ( debug > 2 ) {
                                    dump_rc ( s_box_rc.x[ 0 ], s_box_rc.y[ 0 ], HORI );
                                    dump_rc ( s_box_rc.x[ 0 ], s_box_rc.y[ 0 ], dir );
                                }
                                
                                y_s = s_box_rc.y[0];
                                for ( x_s = 0; x_s < 9; x_s++ ) {
                                    if (( x_s < box_lim[0] ) || ( x_s > box_lim[1] )) {
                                        used_rc_box += elim_possibility ( x_s, y_s, s_box_rc.candidate );
                                    }
                                }
                            }
                        }
                        else {
                            printf( "rc_box(): unknown direction %d (bug)\n", dir );
                        }
    
                    }
                    s_box_rc.total_found = 0;
                }
         
            } while ( next_cell( &x, &y, dir ) == 1 );  
        } // for candidate

    } // for dir
}




void generate_pairs( int max_pair )
{
    struct pair_struct s_pair;

    s_pair.max_pair = max_pair;
  
    for ( s_pair.pair[ 0 ] = 1; s_pair.pair[ 0 ] < 10; s_pair.pair[ 0 ]++ ) {
        for ( s_pair.pair[ 1 ] = s_pair.pair[ 0 ] + 1; s_pair.pair[ 1 ] < 10; s_pair.pair[ 1 ]++ ) {
            if ( max_pair > 2 ) {
                for ( s_pair.pair[ 2 ] = s_pair.pair[ 1 ] + 1; s_pair.pair[ 2 ] < 10; s_pair.pair[ 2 ]++ ) {
                    if ( max_pair > 3 ) {
                        for ( s_pair.pair[ 3 ] = s_pair.pair[ 2 ] + 1; s_pair.pair[ 3 ] < 10; s_pair.pair[ 3 ]++ ) {
                            check_naked_pairs ( s_pair );
                            check_hidden_pairs ( s_pair );
                        }
                    }
                    else {
                        check_naked_pairs ( s_pair );
                        check_hidden_pairs ( s_pair );
                    }
                }
            }
            else {
                check_naked_pairs ( s_pair );
                check_hidden_pairs ( s_pair );
            }
        }
    }
}


/************************************************************************************
 * check_hidden_pairs(): Look for pairs. Stop and eliminate as soon as one is found  *
 ************************************************************************************/
void check_hidden_pairs( struct pair_struct s_pair ) 
{
    int x, y, z, i, dir; 
    int pair_found[ s_pair.max_pair ];
    unsigned int found, found_tmp;

    for ( i = s_pair.max_pair; i < 9; i++ )
        s_pair.pair[i] = -1;

    for ( dir = HORI; dir <= BLOCK; dir++ ) {
        s_pair.total_pairs = 0;
        s_pair.is_naked = -1;  
        s_pair.is_hidden = -1;

        for ( i = 0; i < s_pair.max_pair; i++ ) {
            pair_found[ i ] = 0;
        }
        x = 0;
        y = 0;
        
        do {
            if ( i_puzzle[x][y][0] == 0 ) { 

                found = 0;
                for ( z = 1; (( z < 10 ) && ( i_puzzle[x][y][z] != 0 )); z++ ) {
                    found_tmp = 0;
                    for ( i = 0; i < s_pair.max_pair; i++ ) {
                        if ( i_puzzle[x][y][z] == s_pair.pair[ i ] ) {
                            found_tmp = 1;
                            pair_found[ i ]++;
                        }
                    }
                    if ( found_tmp == 1 ) {
                        found++;
                    }
                    if ( debug > 3 ) {
                        printf ("check_hidden_pairs(): x = %d, y = %d, found_tmp = %d, found = %d, saught = %d\n", x, y, found_tmp, found, i_puzzle[x][y][z]);
                        printf ("check_hidden_pairs(): total_pairs = %d, ", s_pair.total_pairs );
                        for ( i = 0; i < s_pair.max_pair; i++ ) {
                            printf( "pf%d = %d ", i, pair_found[ i ] );
                        }
                        putchar ( '\n' );
                    }
                    
                }
                if ( found > 0 ) {
                    s_pair.x[ s_pair.total_pairs ] = x;
                    s_pair.y[ s_pair.total_pairs ] = y;
                    s_pair.total_pairs++;
                    if ( z > ( found + 1 ) ) {
                        if ( s_pair.is_hidden != 0 ) {
                            s_pair.is_hidden = 1;
                        }
                    }
                }
            }
             
            if ( at_last_element( x, y, dir ) == 1 ) {   
                found = 1;
                for ( i = 0; i < s_pair.max_pair; i++ ){
                    if (( pair_found[ i ] == 0 ) || ( pair_found[ i ] > s_pair.max_pair )) {
                        found = 0;
                    }
                }
                if (( s_pair.total_pairs == s_pair.max_pair ) && ( s_pair.is_hidden == 1 ) && ( found == 1 )) {
                    if ( debug > 1 ) {
                        printf ("We've got hidden %d-pairs at ", s_pair.max_pair );
                    
                        for ( i = 0; i < s_pair.max_pair; i++ ) {
                            printf("(%d, %d) ", s_pair.x[ i ], s_pair.y[ i ]);
                        }
                        printf(" pair =");
                        for ( i = 0; i < s_pair.max_pair; i++ ) {
                            printf(" %d", s_pair.pair[ i ]);
                        }
                        printf(" Direction was %s\n", which_dir( dir ));
                    }
                    delete_pair_elements( s_pair, dir );
                }
                s_pair.is_hidden = -1;
                s_pair.total_pairs = 0;
                for ( i = 0; i < s_pair.max_pair; i++ ) {
                    pair_found[ i ] = 0;
                }
            }
        }  while ( next_cell( &x, &y, dir ) == 1 );  
    }  // for 
}


/************************************************************************************
 * check_naked_pairs(): Look for pairs. Stop and eliminate as soon as one is found  *
 ************************************************************************************/
void check_naked_pairs( struct pair_struct s_pair ) 
{
    int x, y, z, i, dir; 
    int pair_found[ s_pair.max_pair ];
    unsigned int found, found_tmp;

    for ( i = s_pair.max_pair; i < 9; i++ )
        s_pair.pair[i] = -1;

    for ( dir = HORI; dir <= BLOCK; dir++ ) {
        s_pair.total_pairs = 0;
        s_pair.is_naked = -1;  
        s_pair.is_hidden = -1;
        x = 0;
        y = 0;
     
        for ( i = 0; i < s_pair.max_pair; i++ ) {
            pair_found[ i ] = 0;
        }
        
        do {
            
            if ( i_puzzle[x][y][0] == 0 ) { 
                found = 0;
                for ( z = 1; (( z < 10 ) && ( i_puzzle[x][y][z] != 0 )); z++ ) {
                    found_tmp = 0;
                    for ( i = 0; i < s_pair.max_pair; i++ ) {
                        if ( i_puzzle[x][y][z] == s_pair.pair[ i ] ) {
                            pair_found[ i ]++;                     
                            found_tmp = 1;
                        }
                    }
                    if ( debug > 3 ) {
                        printf ("check_naked_pairs(): x = %d, y = %d, found_tmp = %d, found = %d, saught = %d\n", x, y, found_tmp, found, i_puzzle[x][y][z]);
                        printf ("check_naked_pairs(): total_pairs = %d, ", s_pair.total_pairs );
                        for ( i = 0; i < s_pair.max_pair; i++ ) {
                            printf( "pf%d = %d ", i, pair_found[ i ] );
                        }
                        putchar ( '\n' );
                    }
                    if ( found_tmp == 1 ) {
                        found = 1;
                    }
                    else {
                        found = 0;
                        for ( i = 0; i < s_pair.max_pair; i++ ) {
                            if ( i_puzzle[x][y][z] == s_pair.pair[ i ] ) {
                                pair_found[ i ]--;  // undo the element found
                            }
                        }
                        break;
                    }
                }
                    
                if ( found == 1 ) {
                    s_pair.x[ s_pair.total_pairs ] = x;
                    s_pair.y[ s_pair.total_pairs ] = y;
                    s_pair.total_pairs++;
                }
            }
             
            if ( at_last_element( x, y, dir ) == 1 ) {    
                found = 1;
                for ( i = 0; i < s_pair.max_pair; i++ ) {
                    if ( pair_found[ i ] == 0 ) {
                        found = 0;
                    }
                }
                if (( s_pair.total_pairs == s_pair.max_pair ) && ( found == 1 )) {
                    s_pair.is_naked = 1;
                    s_pair.is_hidden = 0;
                    if ( debug > 1 ) {
                        printf ("We've got naked %d-pairs at ", s_pair.max_pair );
                        for ( i = 0; i < s_pair.max_pair; i++ ) {
                            printf("(%d, %d) ", s_pair.x[i], s_pair.y[i]);
                        }
                        printf(" pair =");
                        for ( i = 0; i < s_pair.max_pair; i++ ) {
                            printf(" %d", s_pair.pair[i]);
                        }
                        printf(" Direction was %s\n", which_dir( dir ));
                    }
                    delete_pair_elements( s_pair, dir );
                }
                s_pair.total_pairs = 0;
                s_pair.is_naked = -1;
                for ( i = 0; i < s_pair.max_pair; i++ ) {
                    pair_found[ i ] = 0;
                }
            }
        }  while ( next_cell( &x, &y, dir ) == 1 );  
    }  // for 
}

/*********************************************************************
 * delete_pair_elements() we've got a naked pair, now delete all     *
 * numbers that are in the pair from the other cells in this         *
 * row, column or block                                              *
 * *******************************************************************/


void delete_pair_elements ( struct pair_struct s_pair, int dir )
{
    int x, y, z, i, found, do_delete;
    
    if ( dir == HORI ) {
        x = 0;
        y = s_pair.y[ 0 ];
    }
    else if ( dir == VERT ) {
        x = s_pair.x[ 0 ];
        y = 0;
    }
    else if ( dir == BLOCK ) {  
        x = s_pair.x[ 0 ] - s_pair.x[ 0 ] % 3;
        y = s_pair.y[ 0 ] - s_pair.y[ 0 ] % 3;
    }
    else {
        puts ( "delete_pair_elements(): Unknown direction (bug)" );
    }

    if (( s_pair.is_hidden == 1 ) && ( s_pair.is_naked == 1 )) {
        puts ( "delete_pair_elements(): is_hidden & is_naked (bug)" );
    }
    if (( s_pair.is_hidden == 0 ) && ( s_pair.is_naked == 0 )) {
        puts ( "delete_pair_elements(): !is_hidden & !is_naked (bug)" );
    }

    while (( x < 9 ) && ( y < 9 )) {
        found = 0;
        
        for ( i = 0; i < s_pair.max_pair; i++ ) {
            if (( x == s_pair.x[ i ] ) && ( y == s_pair.y[ i ] )) { // we're on the cell with the pair, if naked, we should not delete anything here. If hidden, we should delete everything but the pair
                found = 1;
            }
        }
        
        if ( found == 0 ) {
            if ( s_pair.is_naked == 1 ) {
                for ( i = 0; i < s_pair.max_pair; i++ ) {   // dumb deletion
                    used_naked_pairs_detection += elim_possibility( x, y, s_pair.pair[ i ] );
                }
            }
        }
        else if ( found == 1 ) {
            if ( s_pair.is_hidden == 1 ) {
                for ( z = 1; (( z < 10 ) && ( i_puzzle[x][y][z] != 0 )); z++ ) {
                    do_delete = 1;
                    for ( i = 0; i < s_pair.max_pair; i++ ) {
                        if ( i_puzzle[x][y][z] == s_pair.pair[ i ] ) {
                            do_delete = 0;  
                        }    
                    }
                    if ( do_delete == 1 ) {     
                        used_hidden_pairs_detection += elim_possibility( x, y, i_puzzle[x][y][z] );
                        z--;  // rewind, we have deleted a possibility
                    }
                }
            }
        }
        
        if ( dir == HORI ) {
            x++;
        }
        else if ( dir == VERT ) {
            y++;    
        }
        else if ( dir == BLOCK ) {
            if (( x % 3 ) == 2 ) {
                if (( y % 3 ) == 2 ) { // we're at the end of the block
                    break;
                }
                else {
                    x -= 2;  // go back to first col of this block
                    y++; // next col
                }
            }
            else {
                x++; // next row
            }
        }
    }
}
   

/*************************************************************
 * elim_possibility(): get rid of a candidate.               *
 *                                                           *
 * return:    0 no candidate removed                         *
 *            1 one cadidate removed                         *
 *************************************************************/

int elim_possibility ( int x, int y, int i_elim )
{
    int z;
    int ret = 0;
    
    for ( z = 1; z < 10; z++ ) {
        if ( i_puzzle[x][y][z] == i_elim ) {
            ret = 1;
            if ( debug > 1 ) {
                printf ("Deleting %d from (%d, %d)\n", i_elim, x, y );
            }
            /* shift array down and fill it up with zeroes at the end */ 
            while ( z < 9 ) {
                i_puzzle[x][y][z] = i_puzzle[x][y][z + 1];  
                z++;
            }
            i_puzzle[x][y][z] = 0;  
            search_determined();
        }
        else if ( i_puzzle[x][y][z] == 0 ) { // no need to search further
            break;
        }
    }
    
    return ( ret );
}


void search_uniques(void)
{
    int i, x_s, y_s, z_s, x_sq, y_sq;
    struct { int x; int y; int i; } s_found;
    
    /* start out with looking for a 
     * number that exists only once in any 3x3 square
     * in the z >= 1 part of the i_puzzle array */
    for ( x_sq = 0; x_sq < 3; x_sq++ ) {
        for ( y_sq = 0; y_sq < 3; y_sq++ ) {   
            for ( i = 1; i < 10; i++ ) {
                s_found.i = 0;
                for ( x_s = ( x_sq * 3 ); x_s < (( (x_sq + 1) * 3 ) ); x_s++ ) {
                    for ( y_s = ( y_sq * 3 ); y_s < (( (y_sq + 1) * 3 ) ); y_s++ ) {
                        for ( z_s = 1; z_s < 10; z_s++ ) {
                            if ( i_puzzle[x_s][y_s][z_s] == 0 ) {
                                break;
                            }
                            else if ( i_puzzle[x_s][y_s][z_s] == i ) {
                                s_found.i++;
                                s_found.x = x_s; // remember the coordinates in case occurence is unique
                                s_found.y = y_s;
                            }
                        }
                    }
                }
                if ( s_found.i == 1 ) { //hey, we found only one occurence, where was it again?
                    i_puzzle[s_found.x][s_found.y][0] = i;
                    for ( s_found.i = 1; s_found.i < 10; s_found.i++ ) { // zero out
                        i_puzzle[s_found.x][s_found.y][s_found.i] = 0;
                    }                    
                    if ( debug > 1 ) {
                        printf("Found a unique number %d on %d, %d\n", i, s_found.x, s_found.y); 
                    }
                    if ( do_show == 1) {
                        print_puzzle();
                    }
                    check_possibilities_s();
                }
            }
        }
    }
}


void search_determined(void) 
{
    int x, y, i_tmp;

    for ( x = 0; x < 9; x++ ) {
        for ( y = 0; y < 9; y++ ) {
            if ( i_puzzle[x][y][0] == 0 ) {
                if ( i_puzzle[x][y][2] == 0 ) { // hey, we have only one number in the z-array. 
                    i_tmp = i_puzzle[x][y][0] = i_puzzle[x][y][1];  
                    i_puzzle[x][y][1] = 0;  
                    if ( debug > 1 ) {
                        printf("Determined a %d at %d, %d\n", i_tmp, x, y);
                    }
                    if ( debug > 3 ) {
                        dump_rc( x, y, HORI );
                    }
                    if ( do_show == 1 ) {
                        print_puzzle();
                    }
                    check_possibilities_s();
                }
            }
        }
    }

}


/*********************************************************
 * puzzle_complete(): check to see if there's any        *
 * unknown cells in the puzzle                           *
 * Returns:                                              *
 *       0: puzzle is incomplete, keep looking           *
 *       1: puzzle is complete                           *
 *********************************************************/


int puzzle_complete(void)
{
    int x, y;
    
    for ( x = 0; x < 9; x++ ) {
        for ( y = 0; y < 9; y++ ) {
            if ( i_puzzle[x][y][0] == 0 ) {
                return(0);
            }
        }
    }
    return(1);
}



int main( int argc, char *argv[] )
{
    struct timeval tv_start_time, tv_cur_time;
    struct timezone tz;
    int i;
    int ten_thousand_loops = 0;
    int j;
    int printed = 0;
    int sec, usec, max_time;
    char str_tot_loops[ 10 ];
    
    if ( argc > 1 ) {
        for ( i = 1; i < argc; i++ ) {
            if ( strncmp( argv[i], "-v", 2 ) == 0 ) {
                printf("ass by Ziad van Beek. Version %s Build %d.\n", VERSION, BUILD );
                return( 0 );
            }
            if ( strncmp( argv[i], "-s", 2 ) == 0 ) {
                debug = 0;
                do_show = 1;        
               // init ncurses
                initscr(); cbreak(); noecho();
                nonl();
                intrflush(stdscr, FALSE);
                keypad(stdscr, TRUE);
                curs_set ( 0 ); // invisible cursor
            }
            if ( strncmp( argv[i], "-d", 2 ) == 0 ) {
                if ( argc > ( i + 1 )) {
                    debug = atoi ( argv[ i+1 ] );
                }
                else {
                    printf("argument -d requires debug level:\n\
                            2 some debug info\n\
                            3 insight on rows, cols and blocks\n\
                            4 provide details about variables\n");
                    return( 0 );
                            
                }
            }
        }
    }   
    
    if ( do_show == 0 ) {
        max_time = 3;
    }
    else {
        max_time = 30;
    }
   
    gettimeofday( &tv_start_time, &tz );
    if ( do_show == 1 ) {
        srand ( tv_start_time.tv_usec );  // seed randomizer for visualisation
    }
    fill_i_puzzle();
    i = 0;

    while ( puzzle_complete() == 0 ) {
        check_possibilities_s();
        search_uniques();
        search_determined();
        x_wing();
        rc_box();
//        swordfish();
        i++;
        // if we seem to bee looping forever, start looking for pairs
        if ( i % 5 == 0 ) {
            gettimeofday( &tv_cur_time, &tz );
            if (( tv_cur_time.tv_sec - tv_start_time.tv_sec ) > max_time ) {
                puts("\nCan't solve this puzzle. Sorry");
                break;   
            }
            for ( j = 2; j < 5; j++ ) {
                generate_pairs( j );
            }
        }
        else {
            if ( i > 10000 ) {
                ten_thousand_loops++;
                i = 0;
            }  
        }   
    }  
    
    gettimeofday( &tv_cur_time, &tz );
    sec = tv_cur_time.tv_sec - tv_start_time.tv_sec;
    if (( tv_cur_time.tv_usec - tv_start_time.tv_usec ) > 0 )
        usec = tv_cur_time.tv_usec - tv_start_time.tv_usec;
    else
        usec = tv_cur_time.tv_usec;
    
    if ( ten_thousand_loops > 0 )
        sprintf (str_tot_loops, "%d%d", ten_thousand_loops, i);
    else
        sprintf (str_tot_loops, "%d", i);
    
    if ( do_show == 1 ) {
        endwin();
        do_show = 0;
    } 
   
    print_puzzle();
    stats( sec, usec, str_tot_loops );

    return(0);
}
